"""
:Copyright:
     Copyright 2015 Lastline, Inc.  All Rights Reserved.
     Created on: Dec 21, 2015 by Kevin Hamacher
"""
from builtins import str  # pylint: disable=redefined-builtin
from builtins import object  # pylint: disable=redefined-builtin
from past.builtins import basestring
import io
import collections
import fs.path
import fs.memoryfs
import os.path
import simplejson
import sys
import tarfile
import time

try:
    import llfile.file

    have_llfile = True
except ImportError:
    have_llfile = False

import llappbundle


def purge_none(dictionary):
    """
    Purge 'None'-entries from a dictionary
    """
    for k in list(dictionary.keys()):
        if dictionary[k] is None:
            del dictionary[k]

    return dictionary


class BundleManipulator(object):
    """
    Allows manipulating or the creation of application bundles
    """

    def __init__(self, logger):
        self._logger = logger

    @staticmethod
    def _strip_metadata_from_tar_object(tar_obj):
        """
        Strips the metadata from a tar object for use as a filter with
        tarfile.add.
        Also used for setting the default values upon manually creating TarInfo
        objects

        :param tar_obj: tarfile.TarInfo
        :returns: stripped tarfile.TarInfo
        """
        # Set the executable bit only if the file is a directory
        # this means the user has the permission to list the content of the dir
        if tar_obj.isdir():
            tar_obj.mode = 0o700
        else:
            tar_obj.mode = 0o600
        tar_obj.uid = 0
        tar_obj.gid = 0
        tar_obj.uname = llappbundle.LLAPP_METADATA_USERNAME
        tar_obj.gname = llappbundle.LLAPP_METADATA_GROUPNAME
        tar_obj.mtime = 0
        return tar_obj

    def create_bundle(self, metadata, files, vfs=None):
        """
        Creates an application bundle with the given files + metadata

        :param metadata: appbundle.Metadata object
        :param files: Dict of filenames, key: full filename, value: relative
                      filename (used in the archive)
        :param vfs: If not null, this fs.* object will be used to retrieve the
                    files instead of opening them on the local file system.
        :returns: io.BytesIO object with the resulting bundle
        """
        self._logger.debug("Creating bundle")
        result = io.BytesIO()
        with tarfile.open(fileobj=result, mode="w", format=tarfile.GNU_FORMAT) as fout:
            # Add metadata first
            metadata = list(metadata.items())
            metadata = collections.OrderedDict(
                sorted(metadata, key=lambda metadata_tuple: metadata_tuple[0])
            )
            metadata_content = simplejson.dumps(purge_none(metadata))
            metadata_tarinfo = tarfile.TarInfo(llappbundle.METADATA_FILENAME)
            metadata_tarinfo.type = tarfile.REGTYPE
            metadata_tarinfo.size = len(metadata_content)
            # Use default (stripped) metadata attributes
            metadata_tarinfo = BundleManipulator._strip_metadata_from_tar_object(metadata_tarinfo)

            fout.addfile(metadata_tarinfo, io.BytesIO(metadata_content.encode("utf-8")))

            # Add regular files
            for fil in sorted(files):
                self._logger.debug("Adding %s", fil)
                if vfs:
                    with vfs.open(str(fil), "rb") as f:
                        file_data = f.read()
                else:
                    with open(fil, "rb") as f:
                        file_data = f.read()

                file_tarinfo = tarfile.TarInfo(
                    fs.path.join(llappbundle.FILES_DIRECTORY, files[fil])
                )
                file_tarinfo.type = tarfile.REGTYPE
                file_tarinfo.size = len(file_data)
                # Use default metadata again
                file_tarinfo = BundleManipulator._strip_metadata_from_tar_object(file_tarinfo)

                fout.addfile(file_tarinfo, io.BytesIO(file_data))

        # Rewind stream
        result.seek(0)
        return result

    def repack_bundle_by_filestream(self, file_handle, metadata=None, archive_password=None):
        """
        Repacks a given bundle or archive specified by the file_handle.
        The file can be password protected, however the archive_password needs
        to be specified.
        See repack_bundle for more details.

        :param file_handle: File handle to the archive
        :param metadata: llappbundle.Metadata object
        :param archive_password: archive password if existing
        :returns: io.BytesIO
        :raises AppBundleNotAnArchiveException: if the given file is not an
                                                archive
        :raises AppBundleInflationException: if the given file can not be
                                             inflated
        """

        # Inflate archive
        assert have_llfile, "Lastline llfile utility required for this functionality"
        ctx = llfile.file.FileContext(name=None, logger=self._logger)
        file_class = llfile.file.File.from_handle(ctx, file_handle)
        if not isinstance(file_class, llfile.file.ArchiveFileInterface):
            raise llappbundle.AppBundleNotAnArchiveException()

        try:
            inflated = file_class.inflate(archive_password)
        except Exception:
            inflated = False

        if not inflated:
            raise llappbundle.AppBundleInflationException()

        return self.repack_bundle(file_class, metadata)

    def repack_bundle(self, file_class, metadata=None):
        """
        Repacks an existing bundle or archive, adding / replacing the metadata
        if the metadata argument is set.

        The resulting hash will always be the same for the same files/metadata

        :param file_class: llfile.file.ArchiveFile to repack, has to be
                           inflated
        :type file_class: `llfile.file.ArchiveFile`
        :param metadata: llappbundle.Metadata or None, overwrites/sets the
                         metadata if specified. If no metadata is specified
                         and the archive does not contain any valid metadata, a
                         AppBundleNoMetadataException is raised.
        :type metadata: `llappbundle.Metadata`
        :returns: io.BytesIO
        :raises AppBundleNoMetadataException: if no valid metadata is specified
                                              and also no metadata is existing
                                              in the existing archive
        :raises AppBundleNotAnArchiveException: if the input file is not an
                                                archive
        :raises AppBundleNotInflatedException: If the input file handle is not
                                               inflated.
        :raises AppBundleFileReadException: Reading the archive content failed
        :raises AppBundleMetadataParsingException: If we are trying to repack a bundle with an
            metadata type that we don't know how to repack
        """
        # Do some magic unpacking here
        assert have_llfile, "Lastline llfile utility required for this functionality"
        if not isinstance(file_class, llfile.file.ArchiveFileInterface):
            raise llappbundle.AppBundleNotAnArchiveException("Can not repack non-archive file")

        if not file_class.is_inflated():
            raise llappbundle.AppBundleNotInflatedException()

        # Unpack files to a memory FS and create a filelist
        memfs = fs.memoryfs.MemoryFS()
        root_dir = "/"
        if metadata and metadata.file_extraction_locations:
            enclosed_files_to_check = metadata.file_extraction_locations
        else:
            enclosed_files_to_check = {fname: fname for fname in file_class.get_names()}
        file_list = self._get_file_list(enclosed_files_to_check, root_dir, file_class, memfs)
        if isinstance(
            metadata, (llappbundle.WindowsBaseMetadata, llappbundle.MacOsExecutableMetadata)
        ):
            metadata = self._normalize_metadata(metadata)
        elif metadata:
            self._logger.debug("Skip normalizing metadata for bundle of type %s", type(metadata))
        else:
            self._logger.debug("Obtaining metadata from extracted data")
            # Obtain metadata from the tempdir
            unpacked_metadata_path = fs.path.join(root_dir, llappbundle.METADATA_FILENAME)
            if memfs.isfile(unpacked_metadata_path):
                with memfs.open(unpacked_metadata_path, "r") as f:
                    content = f.read()
                metadata = llappbundle.parse_metadata(content)
        if not (metadata and metadata.analysis_sample and metadata.analysis_sample.filename):
            raise llappbundle.AppBundleNoMetadataException(
                "Metadata does not exist or does not specify a analysis sample"
            )

        return self.create_bundle(metadata, file_list, memfs)

    def _get_file_list(self, enclosed_files_to_check, root_dir, file_class, memfs):
        """
        Generate the file list for the bundle.

        :param dict enclosed_files_to_check: The files and paths of the archived files.
        :param str root_dir: The root dir inside the memfs.
        :param llfile.file.ArchiveFile file_class: The llfile class instance.
        :param fs.memoryfs.MemoryFS memfs: The memory FS to store the files in

        :returns: The file list dictionary.
        :rtype: dict
        """
        file_list = {}
        for filename, file_details in enclosed_files_to_check.items():
            # NOTE: file_details can be a llappbundle.EnclosedFileDetails or the
            # extraction_location for the file
            filename = self._get_valid_filename(filename)
            full_path = fs.path.join(root_dir, filename)
            dir_name = fs.path.dirname(full_path)
            rel_path = None
            if filename.startswith(llappbundle.FILES_DIRECTORY + "/"):
                rel_path = fs.path.normpath(filename[len(llappbundle.FILES_DIRECTORY) + 1 :])

            self._prepare_memfs_directories(memfs, dir_name)
            self._logger.debug("Unpacking %s", filename)
            # If we have an EnclosedFileDetail, the archive_name can be different than the
            # filename, in case it's not an unicode name, which we still can extract from the
            # llfile archive
            enclosed_name = (
                file_details.archive_name
                if isinstance(file_details, llappbundle.EnclosedFileDetails)
                else filename
            )
            enclosed_file = file_class.get_file(enclosed_name)
            if enclosed_file is None:
                self._logger.warning("Reading %s failed: Cannot access file", enclosed_name)
                raise llappbundle.AppBundleFileReadException("Cannot access file in archive")

            file_data = enclosed_file.get_data()
            if file_data is None:
                self._logger.warning("Reading %s failed: No data returned", enclosed_name)
                raise llappbundle.AppBundleFileReadException("Cannot read file from archive")
            # get the correct data stream (sometimes we get trailing garbage for some reason)
            file_data = file_data[0 : enclosed_file.get_size()]
            with memfs.open(full_path, "wb") as f:
                f.write(file_data)
                # If we're repacking a file with the correct layout, remove the
                # FILES_DIRECTORY prefix since it will be re-added in the
                # creation process
                if rel_path:
                    file_list[full_path] = rel_path
                # If we're adding files with a name different than
                # METADATA_FILENAME, move the files to the FILES_DIRECTORY
                # folder in order to generate a valid application bundle
                elif filename != llappbundle.METADATA_FILENAME:
                    file_list[full_path] = fs.path.normpath(filename)
        return file_list

    def _prepare_memfs_directories(self, memfs, dir_name):
        """
        Create the directories in the memfs leading to the final destination.

        :param fs.memoryfs.MemoryFS memfs: The memfs instance.
        :param str dirname: The full directory path.
        """
        if not memfs.exists(dir_name) and dir_name != "":
            cur_path = ""
            for cur_dir in dir_name.strip().split("/"):
                if cur_dir == "":
                    continue
                cur_path = os.path.join(cur_path, cur_dir)
                if not memfs.exists(cur_path):
                    memfs.makedir(cur_path)

    def _get_valid_filename(self, filename):
        """
        Get a valid encoded filename from the provided filename.

        :param str filename: The filename input.

        :return: The valid filename.
        :rtype: str

        :raises: llappbundle.AppBundleFileReadException
        """
        if sys.version_info.major < 3 and isinstance(filename, basestring):
            try:
                # This is the name we will use when creating the bundle. Note that this is not
                # the extraction location but rather what name we will use in the file
                # when creating the new bundle
                return filename.decode("utf8")
            except UnicodeError:
                raise llappbundle.AppBundleFileReadException("Non UTF filename in bundle")
        return filename

    def _normalize_metadata(self, metadata):
        """
        Normalize the metadata for WindowsBaseMetadata and llappbundle.MacOsExecutableMetadata.

        :param llappbundle.WindowsBaseMetadata|llappbundle.MacOsExecutableMetadata metadata:
            The metadate to normalize.

        :returns: The normalized metadata
        :rtype: llappbundle.WindowsExecutableMetadata|llappbundle.MacOsExecutableMetadata|
            llappbundle.DocumentMetadata
        """
        new_analysis_sample = None
        new_file_extraction_locations = None
        # Normalize analysis_sample and file_extraction_locations
        analysis_sample = metadata.analysis_sample
        if analysis_sample:
            new_analysis_sample = llappbundle.Command(
                filename=fs.path.normpath(analysis_sample.filename)
                if analysis_sample.filename
                else None,
                executable=fs.path.normpath(analysis_sample.executable)
                if analysis_sample.executable
                else None,
                parameters=analysis_sample.parameters,
                run_directory=fs.path.normpath(analysis_sample.run_directory)
                if analysis_sample.run_directory
                else None,
            )
        # We need to create the new file_extraction_locations, since now archive_name and
        # filename are the same
        if metadata.file_extraction_locations:
            new_file_extraction_locations = {
                fs.path.normpath(k): fs.path.normpath(llappbundle.get_extraction_location(v))
                for k, v in metadata.file_extraction_locations.items()
            }

        if isinstance(metadata, llappbundle.WindowsExecutableMetadata):
            metadata = llappbundle.WindowsExecutableMetadata(
                analysis_environment=metadata.analysis_environment,
                analysis_sample=new_analysis_sample,
                os_environment_values=metadata.os_environment_values,
                preparation_commands=metadata.preparation_commands,
                registry_values=metadata.registry_values,
                file_extraction_locations=new_file_extraction_locations,
                allow_rename=metadata.allow_rename,
            )
        elif isinstance(metadata, llappbundle.MacOsExecutableMetadata):
            metadata = llappbundle.MacOsExecutableMetadata(
                analysis_environment=metadata.analysis_environment,
                analysis_sample=new_analysis_sample,
                os_environment_values=metadata.os_environment_values,
                preparation_commands=metadata.preparation_commands,
                file_extraction_locations=new_file_extraction_locations,
                allow_rename=metadata.allow_rename,
            )
        elif isinstance(metadata, llappbundle.DocumentMetadata):
            metadata = llappbundle.DocumentMetadata(
                analysis_environment=metadata.analysis_environment,
                analysis_sample=new_analysis_sample,
                os_environment_values=metadata.os_environment_values,
                registry_values=metadata.registry_values,
                file_extraction_locations=new_file_extraction_locations,
                passwords=metadata.passwords,
            )
        return metadata

    def verify_metadata(self, metadata, file_list, strip_files_prefix=False):
        """
        Verify that the given metadata contains all required values.
        Also logs if files mentioned in the metadata are missing in :file_list

        :param metadata: metadata object to be verified
        :type metadata: llappbundle.Metadata
        :param list(str) file_list: relative filenames in the archive without the
            llappbundle_files/ prefix if strip_files_prefix=False
        :param bool strip_files_prefix: Strips the llappbundle_files/ prefix from
            the files. Files not starting with the prefix will be ignored!
        :raises AppBundleInvalidAppBundleException: If verification failed
        """
        # List of files that are specified in the metadata and have to be
        # existing:
        required_files = []

        # Verify that the analysis_sample entry is existing
        analysis_sample = metadata.analysis_sample
        if not analysis_sample or not analysis_sample.filename:
            raise llappbundle.AppBundleInvalidAppBundleException("analysis sample not specified")

        required_files.append(analysis_sample.filename)

        # Check preparation_commands
        if metadata.preparation_commands:
            for prep_cmd in metadata.preparation_commands:
                if not prep_cmd.filename and not prep_cmd.executable:
                    raise llappbundle.AppBundleInvalidAppBundleException(
                        "file not specified for a preperation command"
                    )

                required_files.append(prep_cmd.filename)

        if not metadata.file_extraction_locations:
            raise llappbundle.AppBundleInvalidAppBundleException(
                "no file extraction locations given"
            )

        # Map required files to archive files using file_extraction_locations
        for dest in metadata.file_extraction_locations.values():
            dest = llappbundle.get_extraction_location(dest)
            if dest in required_files:
                required_files.remove(dest)

        # These are only warnings since the file might execute a builtin
        # windows executable
        self._logger.debug("Validating existance of required files")
        for req_file in required_files:
            self._logger.info(
                "Could not find program %s in the archive, assuming that "
                "it already exists in the OS %%PATH%%.",
                req_file,
            )
        self._check_file_list(file_list, strip_files_prefix, metadata)

    def _check_file_list(self, file_list, strip_files_prefix, metadata):
        """
        Check the file list for being complete and strip prefixes if configured so.

        :param list[str] file_list: The list of files.
        :param bool strip_files_prefix: If the llappbundle prefix should be stripped.
        :param llappbundle.Metadata metadata: The Appbundle metadata.
        """
        if not file_list:
            self._logger.debug("No filelist specified, skipping validation of required files")
            return
        if strip_files_prefix:
            file_list = [
                name[len(llappbundle.FILES_DIRECTORY) + 1 :]
                for name in file_list
                if name.startswith(llappbundle.FILES_DIRECTORY)
            ]

        for req_file in list(metadata.file_extraction_locations.keys()):
            if req_file not in file_list:
                self._logger.warning(
                    "Could not find file %s in the archive (supposed to be copied)", req_file
                )
                raise llappbundle.AppBundleInvalidAppBundleException(
                    "Required file not found in the archive"
                )

    def inflate_appbundle(self, appbundle, extract_dir):
        """
        Inflates the appbundle to the given path, returning the metadata

        :param appbundle: appbundle file object
        :type appbundle: file
        :param extract_dir: Path to extract to
        :type extract_dir: str
        :raises AppBundleInflationException: if the given file can not be inflated
        :raises AppBundleInvalidAppBundleException: If appbundle verification fails
        :returns: The metadata object
        :rtype: Metadata
        """
        filelist = []
        try:
            tar_file = tarfile.open(fileobj=appbundle, mode="r")
            with tar_file as tar:
                # Build a filelist without the llappbundle_files/ prefix
                filelist = [
                    m.name[len(llappbundle.FILES_DIRECTORY) + 1 :]
                    for m in tar.members
                    if m.name.startswith(llappbundle.FILES_DIRECTORY)
                ]
                for tar_member in tar.getmembers():
                    if not tar_member.mtime:
                        self._logger.info("Changing timestamp of %s to now", tar_member.name)
                        tar_member.mtime = time.time()
                    tar.extract(tar_member, path=extract_dir)
        except Exception as e:
            self._logger.exception("Extracting appbundle failed: %s", e)
            raise llappbundle.AppBundleInflationException()

        try:
            with open(fs.path.join(extract_dir, llappbundle.METADATA_FILENAME), "r") as f:
                metadata = llappbundle.parse_metadata(f.read())
        except IOError as e:
            self._logger.warning("Metadata file does not exist: %s", e)
            raise llappbundle.AppBundleInflationException()

        self.verify_metadata(metadata, filelist)

        return metadata

    def get_metadata_file(self, appbundle):
        """
        Retrieves the metadata from an appbundle archive.

        :param appbundle: appbundle file object
        :type appbundle: file
        :returns: The metadata object
        :rtype: Metadata
        """
        try:
            with tarfile.open(fileobj=appbundle, mode="r") as tar:
                metadata_file = tar.extractfile(llappbundle.METADATA_FILENAME)
        except Exception as e:
            self._logger.exception("Extracting metadata from appbundle failed: %s", e)
            raise llappbundle.AppBundleInflationException()

        try:
            metadata = llappbundle.parse_metadata(metadata_file.read())
        except IOError as e:
            self._logger.warning("Metadata file does not exist: %s", e)
            raise llappbundle.AppBundleInflationException()

        self.verify_metadata(metadata, None)

        return metadata
