"""
:Copyright:
     Copyright 2016 Lastline, Inc.  All Rights Reserved.
     Created on: Apr 21, 2016 by Kevin Hamacher

Helper functions to create appbundles from the python command line
"""
from builtins import str  # pylint: disable=redefined-builtin
from past.builtins import basestring
import collections
import fs.memoryfs
import sys

try:
    from llutils import lllog

    have_lllog = True
except ImportError:
    import logging

    have_lllog = False

import llappbundle
import llappbundle.bundle


def _is_valid_utf8(s):
    if sys.version_info.major < 3 and isinstance(s, bytes):
        try:
            s.decode("utf-8")
        except UnicodeDecodeError:
            return False
    return isinstance(s, basestring)


FileInfos = collections.namedtuple(
    "FileInfos", ["file_extraction_locations", "file_names", "archive_files", "memfs"]
)


def _prepare_bundle_files(files, main_subject, logger):
    """
    Prepare the files of a bundle

    :param files: Files to be stored in the appbundle
    :type files: dict(virtual filename => stream)
    :param main_subject: Virtual filename of the analysis subject
    :type main_subject: str
    :param logger: logger used for logging
    :type logger: logger
    :returns: Infos about the files in the bundle
    :rtype: FileInfos
    :raises: llappbundle.AppBundleInvalidAppBundleException if the specified
             appbundle-parameters are invalid.
    """
    if main_subject not in files:
        logger.error("Main subject was not found the supplied files.")
        raise llappbundle.AppBundleInvalidAppBundleException(
            "Main subject is not inside of the appbundle."
        )

    if not _is_valid_utf8(main_subject):
        logger.error("Main subject must be valid utf-8: %s", main_subject)
        raise llappbundle.AppBundleInvalidAppBundleException(
            "Main subject must be valid utf-8: %s" % main_subject
        )

    memfs = fs.memoryfs.MemoryFS()

    # Generate the archive files + file extraction locations
    file_extraction_locations = {}
    file_names = []
    archive_files = {}

    file_nr = 0
    for virt_fn, filestream in files.items():
        if not _is_valid_utf8(virt_fn):
            logger.warning("Virtual file name is not valid utf-8, skipped: %s", virt_fn)
            continue
        # Generate name
        archive_file_name = "file_%d" % file_nr
        # Store in memfs
        old_pos = filestream.tell()
        with memfs.open(str(archive_file_name), "wb") as f:
            f.write(filestream.read())
        filestream.seek(old_pos)
        # Add to files to extract
        file_extraction_locations[archive_file_name] = virt_fn
        file_names.append(archive_file_name)
        archive_files[archive_file_name] = archive_file_name
        file_nr += 1

    return FileInfos(
        file_extraction_locations=file_extraction_locations,
        file_names=file_names,
        archive_files=archive_files,
        memfs=memfs,
    )


def _internal_create_appbundle(file_infos, metadata, logger):
    """
    Creates an Lastline Application Bundle.

    :param file_infos: A FileInfos with information about the bundle files
    :type file_infos: `FileInfos`
    :param metadata: A metadata information about the bundle that will be creted
    :type metadata: `llappbundle.MetadataBase`
    :param logger: logger used for logging
    :type logger: logger
    :returns: appbundle
    :rtype: BytesIO
    :raises: llappbundle.AppBundleInvalidAppBundleException if the specified
             appbundle-parameters are invalid.
    """
    bundle_manipulator = llappbundle.bundle.BundleManipulator(logger)

    try:
        bundle_manipulator.verify_metadata(metadata, file_infos.file_names)
    except llappbundle.AppBundleInvalidAppBundleException as e:
        logger.exception("Verifying metadata failed with an exception: %s", e)
        raise

    return bundle_manipulator.create_bundle(metadata, file_infos.archive_files, file_infos.memfs)


def create_windows_appbundle(
    files,
    main_subject,
    executable=None,
    arguments=None,
    logger=None,
    run_directory=None,
    registry_values=None,
    allow_rename=False,
    preparation_commands=None,
):
    """
    Creates a Lastline Application Bundle for Windows environment.

    :param dict(virtual filename => stream) files: Files to be stored in the appbundle
    :param str main_subject: Virtual filename of the analysis subject
    :param str executable: Can be used to override the executable. Setting this
                       will disable the detection of how to execute the main
                       subject.
    :param list(str) arguments: arguments that should be provided to the executable if set.
    :param str run_directory: Directory in which the subject should be started in.
    :param logging.Logger logger: logger used for logging
    :param dict registry_values: A dictionary with registry values to be added in the bundle
        metadata
    :param bool allow_rename: A bool that indicate if the analysis subject can be renamed or not
    :param list(llappbundle.Command) preparation_commands: A list with commands to be executed
        in the system before running the sample
    :returns: appbundle
    :rtype: BytesIO
    :raises: llappbundle.AppBundleInvalidAppBundleException if the specified
             appbundle-parameters are invalid.
    """
    if not logger and have_lllog:
        logger = lllog.SimpleLogger.get_simple_logger(None, None)
    elif not logger:
        logger = logging

    file_infos = _prepare_bundle_files(files, main_subject, logger)

    arguments = arguments or []
    analysis_sample = llappbundle.Command(
        filename=main_subject,
        parameters=arguments,
        run_directory=run_directory,
        executable=executable,
    )

    metadata = llappbundle.WindowsExecutableMetadata(
        analysis_environment=None,
        analysis_sample=analysis_sample,
        os_environment_values=None,
        preparation_commands=preparation_commands,
        registry_values=registry_values,
        file_extraction_locations=file_infos.file_extraction_locations,
        allow_rename=allow_rename,
    )
    return _internal_create_appbundle(file_infos=file_infos, metadata=metadata, logger=logger)


# For legacy compatibility
create_appbundle = create_windows_appbundle
internal_create_appbundle = create_windows_appbundle


def create_macos_appbundle(
    files,
    main_subject,
    executable=None,
    arguments=None,
    logger=None,
    run_directory=None,
    allow_rename=False,
    preparation_commands=None,
):
    """
    Creates a Lastline Application Bundle for a MacOs environment.

    :param dict(virtual filename => stream) files: Files to be stored in the appbundle
    :param str main_subject: Virtual filename of the analysis subject
    :param str executable: Can be used to override the executable. Setting this will disable the
        detection of how to execute the main subject.
    :param list(str) arguments: arguments that should be provided to the executable if set.
    :param logging.Logger logger: logger used for logging
    :param str run_directory: Directory in which the subject should be started in.
    :param bool allow_rename: A bool that indicate if the analysis subject can be renamed or not
    :param list(llappbundle.Command) preparation_commands: A list with commands to be executed
        in the system before running the sample
    :returns: appbundle
    :rtype: BytesIO
    :raises: llappbundle.AppBundleInvalidAppBundleException if the specified appbundle-parameters
        are invalid.
    """
    if not logger and have_lllog:
        logger = lllog.SimpleLogger.get_simple_logger(None, None)
    elif not logger:
        logger = logging

    file_infos = _prepare_bundle_files(files, main_subject, logger)

    arguments = arguments or []
    analysis_sample = llappbundle.Command(
        filename=main_subject,
        parameters=arguments,
        run_directory=run_directory,
        executable=executable,
    )

    metadata = llappbundle.MacOsExecutableMetadata(
        analysis_environment=None,
        analysis_sample=analysis_sample,
        os_environment_values=None,
        preparation_commands=preparation_commands,
        file_extraction_locations=file_infos.file_extraction_locations,
        allow_rename=allow_rename,
    )
    return _internal_create_appbundle(file_infos=file_infos, metadata=metadata, logger=logger)


def create_document_bundle(files, main_subject, logger=None, passwords=None):
    """
    Creates an Lastline Application Bundle for documents

    :param dict(virtual filename => stream) files: Files to be stored in the appbundle
    :param str main_subject: Virtual filename of the analysis subject
    :param logging.Logger logger: logger used for logging
    :param list(str)passwords: Passwords to add to the metadata for decryption later
    :returns: appbundle
    :rtype: BytesIO
    :raises: llappbundle.AppBundleInvalidAppBundleException if the specified
             appbundle-parameters are invalid.
    """
    if not logger and have_lllog:
        logger = lllog.SimpleLogger.get_simple_logger(None, None)
    elif not logger:
        logger = logging

    file_infos = _prepare_bundle_files(files, main_subject, logger)

    analysis_sample = llappbundle.Command(
        filename=main_subject, parameters=None, run_directory=None, executable=None
    )

    metadata = llappbundle.DocumentMetadata(
        analysis_environment=None,
        analysis_sample=analysis_sample,
        os_environment_values=None,
        registry_values=None,
        file_extraction_locations=file_infos.file_extraction_locations,
        passwords=passwords,
    )

    bundle_manipulator = llappbundle.bundle.BundleManipulator(logger)

    try:
        bundle_manipulator.verify_metadata(metadata, file_infos.file_names)
    except llappbundle.AppBundleInvalidAppBundleException as e:
        logger.exception("Verifying metadata failed with an exception: %s", e)
        raise

    return bundle_manipulator.create_bundle(metadata, file_infos.archive_files, file_infos.memfs)


def create_web_replay_bundle_with_file(files, pcap, main_subject, logger=None):
    """
    Creates an Lastline Application Bundle for Web Replay with a file subject.

    Web Replay analyses allow to analyze a URL, HTML or JavaScript file together with a pcap. The
    traffic inside the pcap is matched on the network requests during the analysis. Note, that only
    content of the pcap will be served. No requests to the internet will be made.

    :param dict(virtual filename => stream) files: Files to be stored in the appbundle
    :param str pcap: Virtual filename of the pcap in the files dictionary
    :param str main_subject: Virtual filename of the file analysis subject
    :param logging.Logger logger: logger used for logging
    :returns: appbundle
    :rtype: BytesIO
    :raises: llappbundle.AppBundleInvalidAppBundleException if the specified
             appbundle-parameters are invalid.
    """
    if not logger and have_lllog:
        logger = lllog.SimpleLogger.get_simple_logger(None, None)
    elif not logger:
        logger = logging

    if main_subject not in files:
        logger.error("Main subject was not found in the supplied files.")
        raise llappbundle.AppBundleInvalidAppBundleException(
            "Main subject is not inside of the appbundle."
        )

    file_infos = _prepare_bundle_files(files, pcap, logger)

    analysis_sample = llappbundle.WebReplayCommand(filename=pcap, subject=main_subject)

    metadata = llappbundle.WebReplayMetadata(
        analysis_environment=None,
        analysis_sample=analysis_sample,
        os_environment_values=None,
        registry_values=None,
        file_extraction_locations=file_infos.file_extraction_locations,
    )

    bundle_manipulator = llappbundle.bundle.BundleManipulator(logger)

    try:
        bundle_manipulator.verify_metadata(metadata, file_infos.file_names)
    except llappbundle.AppBundleInvalidAppBundleException as e:
        logger.exception("Verifying metadata failed with an exception: %s", e)
        raise

    return bundle_manipulator.create_bundle(metadata, file_infos.archive_files, file_infos.memfs)


def create_web_replay_bundle_with_url(files, pcap, main_subject, logger=None):
    """
    Creates an Lastline Application Bundle for Web Replay with an URL subject.

    Web Replay analyses allow to analyze a URL, HTML or JavaScript file together with a pcap. The
    traffic inside the pcap is matched on the network requests during the analysis. Note, that only
    content of the pcap will be served. No requests to the internet will be made.

    :param dict(virtual filename => stream) files: Files to be stored in the appbundle
    :param str pcap: Virtual filename of the pcap in the files dictionary
    :param str main_subject: URL Analysis subject
    :param logging.Logger logger: logger used for logging
    :returns: appbundle
    :rtype: BytesIO
    :raises: llappbundle.AppBundleInvalidAppBundleException if the specified
             appbundle-parameters are invalid.
    """
    if not logger and have_lllog:
        logger = lllog.SimpleLogger.get_simple_logger(None, None)
    elif not logger:
        logger = logging

    file_infos = _prepare_bundle_files(files, pcap, logger)

    analysis_sample = llappbundle.WebReplayCommand(filename=pcap, subject=main_subject)

    metadata = llappbundle.WebReplayMetadata(
        analysis_environment=None,
        analysis_sample=analysis_sample,
        os_environment_values=None,
        registry_values=None,
        file_extraction_locations=file_infos.file_extraction_locations,
    )

    bundle_manipulator = llappbundle.bundle.BundleManipulator(logger)

    try:
        bundle_manipulator.verify_metadata(metadata, file_infos.file_names)
    except llappbundle.AppBundleInvalidAppBundleException as e:
        logger.exception("Verifying metadata failed with an exception: %s", e)
        raise

    return bundle_manipulator.create_bundle(metadata, file_infos.archive_files, file_infos.memfs)
